@@ -1,5 +1,5 @@
 /* darkstat 3
- * copyright (c) 2001-2007 Emil Mikulic.
+ * copyright (c) 2001-2008 Emil Mikulic.
  *
  * cap.c: interface to libpcap.
  *
@@ -51,6 +51,7 @@ cap_init(const char *device, const char *filter, int promisc)
 {
    char errbuf[PCAP_ERRBUF_SIZE], *tmp_device;
    int linktype, caplen;
+   extern int want_pppoe;
 
    /* pcap doesn't like device being const */
    tmp_device = xstrdup(device);
@@ -69,15 +70,21 @@ cap_init(const char *device, const char *filter, int promisc)
 
    /* Work out the linktype and what caplen it needs. */
    linktype = pcap_datalink(pcap);
+   verbosef("linktype is %d", linktype);
+   if (linktype == DLT_EN10MB)
+      show_mac_addrs = 1;
    linkhdr = getlinkhdr(linktype);
    if (linkhdr == NULL)
       errx(1, "unknown linktype %d", linktype);
    if (linkhdr->handler == NULL)
       errx(1, "no handler for linktype %d", linktype);
    caplen = getcaplen(linkhdr);
+   if (want_pppoe) {
+      caplen += PPPOE_HDR_LEN;
+      if (linktype != DLT_EN10MB)
+         errx(1, "can't do PPPoE decoding on a non-Ethernet linktype");
+   }
    verbosef("caplen is %d", caplen);
-   if (linktype == DLT_EN10MB)
-      show_mac_addrs = 1;
 
    /* Close and re-open pcap to use the new caplen. */
    pcap_close(pcap);
@@ -233,16 +240,72 @@ cap_poll(fd_set *read_set
       break;
 #endif
    }
-   /*FIXME*/fprintf(stderr, "%-20d\r", total);
+   /*FIXME*/if (want_verbose) fprintf(stderr, "%-20d\r", total);
 }
 
-/*
- * Stop pcap.
- */
 void
 cap_stop(void)
 {
    pcap_close(pcap);
 }
 
+/* Run through entire capfile. */
+void
+cap_from_file(const char *capfile, const char *filter)
+{
+   char errbuf[PCAP_ERRBUF_SIZE];
+   int linktype, ret;
+
+   /* Open packet capture descriptor. */
+   errbuf[0] = '\0'; /* zero length string */
+   pcap = pcap_open_offline(capfile, errbuf);
+
+   if (pcap == NULL)
+      errx(1, "pcap_open_offline(): %s", errbuf);
+
+   if (errbuf[0] != '\0') /* not zero length anymore -> warning */
+      warnx("pcap_open_offline() warning: %s", errbuf);
+
+   /* Work out the linktype. */
+   linktype = pcap_datalink(pcap);
+   linkhdr = getlinkhdr(linktype);
+   if (linkhdr == NULL)
+      errx(1, "unknown linktype %d", linktype);
+   if (linkhdr->handler == NULL)
+      errx(1, "no handler for linktype %d", linktype);
+   if (linktype == DLT_EN10MB) /* FIXME: impossible with capfile? */
+      show_mac_addrs = 1;
+
+   /* Set filter expression, if any. */ /* FIXME: factor! */
+   if (filter != NULL)
+   {
+      struct bpf_program prog;
+      char *tmp_filter = xstrdup(filter);
+      if (pcap_compile(
+            pcap,
+            &prog,
+            tmp_filter,
+            1,          /* optimize */
+            0)          /* netmask */
+            == -1)
+         errx(1, "pcap_compile(): %s", pcap_geterr(pcap));
+
+      if (pcap_setfilter(pcap, &prog) == -1)
+         errx(1, "pcap_setfilter(): %s", pcap_geterr(pcap));
+
+      pcap_freecode(&prog);
+      free(tmp_filter);
+   }
+
+   /* Process file. */
+   ret = pcap_dispatch(
+         pcap,
+         -1,               /* count, -1 = entire buffer */
+         linkhdr->handler, /* callback func from decode.c */
+         NULL);            /* user */
+
+   if (ret < 0)
+      errx(1, "pcap_dispatch(): %s", pcap_geterr(pcap));
+}
+
 /* vim:set ts=3 sw=3 tw=78 expandtab: */
