@@ -1,5 +1,5 @@
 /* darkstat 3
- * copyright (c) 2001-2007 Emil Mikulic.
+ * copyright (c) 2001-2008 Emil Mikulic.
  *
  * darkstat.c: signals, cmdline parsing, program body.
  *
@@ -46,11 +46,36 @@ static volatile int reset_pending = 0;
 static void sig_reset(int signum _unused_) { reset_pending = 1; }
 
 /* --- Commandline parsing --- */
+static unsigned long
+parsenum(const char *str, unsigned long max /* 0 for no max */)
+{
+   unsigned long n;
+   char *end;
+
+   errno = 0;
+   n = strtoul(str, &end, 10);
+   if (*end != '\0')
+      errx(1, "\"%s\" is not a valid number", str);
+   if (errno == ERANGE)
+      errx(1, "\"%s\" is out of range", str);
+   if ((max != 0) && (n > max))
+      errx(1, "\"%s\" is out of range (max %lu)", str, max);
+   return n;
+}
+
 const char *interface = NULL;
 static void cb_interface(const char *arg) { interface = arg; }
 
+const char *capfile = NULL;
+static void cb_capfile(const char *arg) { capfile = arg; }
+
+int want_pppoe = 0;
+static void cb_pppoe(const char *arg _unused_) { want_pppoe = 1; }
+
+static void cb_verbose(const char *arg _unused_) { want_verbose = 1; }
+
 int want_daemonize = 1;
-static void cb_debug(const char *arg _unused_) { want_daemonize = 0; }
+static void cb_no_daemon(const char *arg _unused_) { want_daemonize = 0; }
 
 int want_promisc = 1;
 static void cb_no_promisc(const char *arg _unused_) { want_promisc = 0; }
@@ -59,19 +84,7 @@ int want_dns = 1;
 static void cb_no_dns(const char *arg _unused_) { want_dns = 0; }
 
 unsigned short bindport = 667;
-static void cb_port(const char *arg)
-{
-   unsigned long p;
-   char *ep;
-
-   errno = 0;
-   p = strtoul(arg, &ep, 10);
-   if (*ep != '\0')
-      errx(1, "\"%s\" is not a valid number", arg);
-   if ((errno == ERANGE) || (p > 65535))
-      errx(1, "\"%s\" is out of range", arg);
-   bindport = (unsigned short)p;
-}
+static void cb_port(const char *arg) { bindport = parsenum(arg, 65536); }
 
 in_addr_t bindaddr = INADDR_ANY;
 static void cb_bindaddr(const char *arg)
@@ -115,7 +128,7 @@ static void cb_import(const char *arg)
 const char *export_fn = NULL;
 static void cb_export(const char *arg)
 {
-   if (chroot_dir == NULL)
+   if ((chroot_dir == NULL) && (capfile == NULL))
       errx(1, "the export file is relative to the chroot.\n"
       "You must specify a --chroot dir before you can use --export.");
    else
@@ -132,31 +145,58 @@ static void cb_pidfile(const char *arg)
       pid_fn = arg;
 }
 
+unsigned int hosts_max = 1000;
+static void cb_hosts_max(const char *arg)
+{ hosts_max = parsenum(arg, 0); }
+
+unsigned int hosts_keep = 500;
+static void cb_hosts_keep(const char *arg)
+{ hosts_keep = parsenum(arg, 0); }
+
+unsigned int ports_max = 200;
+static void cb_ports_max(const char *arg)
+{ ports_max = parsenum(arg, 65536); }
+
+unsigned int ports_keep = 30;
+static void cb_ports_keep(const char *arg)
+{ ports_keep = parsenum(arg, 65536); }
+
+unsigned int highest_port = 65535;
+static void cb_highest_port(const char *arg)
+{ highest_port = parsenum(arg, 65535); }
+
 /* --- */
 
 struct cmdline_arg {
-   const char *name, *arg_name; /* NULL arg_name means no arg */
-   const enum { MANDATORY, OPTIONAL } need;
+   const char *name, *arg_name; /* NULL arg_name means unary */
    void (*callback)(const char *arg);
    int num_seen;
 };
 
 static struct cmdline_arg cmdline_args[] = {
-   {"-i",           "interface",       MANDATORY, cb_interface,  0},
-   {"--debug",      NULL,              OPTIONAL,  cb_debug,      0},
-   {"--no-promisc", NULL,              OPTIONAL,  cb_no_promisc, 0},
-   {"--no-dns",     NULL,              OPTIONAL,  cb_no_dns,     0},
-   {"-p",           "port",            OPTIONAL,  cb_port,       0},
-   {"-b",           "bindaddr",        OPTIONAL,  cb_bindaddr,   0},
-   {"-f",           "filter",          OPTIONAL,  cb_filter,     0},
-   {"-l",           "network/netmask", OPTIONAL,  cb_local,      0},
-   {"--chroot",     "dir",             OPTIONAL,  cb_chroot,     0},
-   {"--user",       "username",        OPTIONAL,  cb_user,       0},
-   {"--daylog",     "filename",        OPTIONAL,  cb_daylog,     0},
-   {"--import",     "filename",        OPTIONAL,  cb_import,     0},
-   {"--export",     "filename",        OPTIONAL,  cb_export,     0},
-   {"--pidfile",    "filename",        OPTIONAL,  cb_pidfile,    0},
-   {NULL,           NULL,              OPTIONAL,  NULL,          0}
+   {"-i",             "interface",       cb_interface,    0},
+   {"-r",             "file",            cb_capfile,      0},
+   {"--pppoe",        NULL,              cb_pppoe,        0},
+   {"--verbose",      NULL,              cb_verbose,      0},
+   {"--no-daemon",    NULL,              cb_no_daemon,    0},
+   {"--no-promisc",   NULL,              cb_no_promisc,   0},
+   {"--no-dns",       NULL,              cb_no_dns,       0},
+   {"-p",             "port",            cb_port,         0},
+   {"-b",             "bindaddr",        cb_bindaddr,     0},
+   {"-f",             "filter",          cb_filter,       0},
+   {"-l",             "network/netmask", cb_local,        0},
+   {"--chroot",       "dir",             cb_chroot,       0},
+   {"--user",         "username",        cb_user,         0},
+   {"--daylog",       "filename",        cb_daylog,       0},
+   {"--import",       "filename",        cb_import,       0},
+   {"--export",       "filename",        cb_export,       0},
+   {"--pidfile",      "filename",        cb_pidfile,      0},
+   {"--hosts-max",    "count",           cb_hosts_max,    0},
+   {"--hosts-keep",   "count",           cb_hosts_keep,   0},
+   {"--ports-max",    "count",           cb_ports_max,    0},
+   {"--ports-keep",   "count",           cb_ports_keep,   0},
+   {"--highest-port", "port",            cb_highest_port, 0},
+   {NULL,             NULL,              NULL,            0}
 };
 
 static void
@@ -175,17 +215,17 @@ usage(void)
    int width, first;
    struct cmdline_arg *arg;
 
-   printf("\n");
+   printf(PACKAGE_STRING " (built with libpcap %d.%d)\n\n",
+      PCAP_VERSION_MAJOR, PCAP_VERSION_MINOR);
+
    width = printf("usage: darkstat ");
    first = 1;
 
    for (arg = cmdline_args; arg->name != NULL; arg++) {
       if (first) first = 0; else pad(width);
-      if (arg->need == OPTIONAL) printf("[ ");
-      printf("%s", arg->name);
+      printf("[ %s", arg->name);
       if (arg->arg_name != NULL) printf(" %s", arg->arg_name);
-      if (arg->need == OPTIONAL) printf(" ]");
-      printf("\n");
+      printf(" ]\n");
    }
    printf("\n"
 "Please refer to the darkstat(1) manual page for further\n"
@@ -232,8 +272,6 @@ parse_sub_cmdline(const int argc, char * const *argv)
 static void
 parse_cmdline(const int argc, char * const *argv)
 {
-   struct cmdline_arg *arg;
-
    if (argc < 1) {
       /* Not enough args. */
       usage();
@@ -242,17 +280,38 @@ parse_cmdline(const int argc, char * const *argv)
 
    parse_sub_cmdline(argc, argv);
 
-   for (arg = cmdline_args; arg->name != NULL; arg++)
-      if ((arg->need == MANDATORY) && (arg->num_seen == 0)) {
-         printf("\nerror: failed to specify mandatory argument \"%s\"\n",
-            arg->name);
-         usage();
-         exit(EXIT_FAILURE);
-      }
-
    /* some default values */
    if (chroot_dir == NULL) chroot_dir = CHROOT_DIR;
    if (privdrop_user == NULL) privdrop_user = PRIVDROP_USER;
+
+   /* sanity check args */
+   if ((interface == NULL) && (capfile == NULL))
+      errx(1, "must specify either interface (-i) or capture file (-r)");
+
+   if ((interface != NULL) && (capfile != NULL))
+      errx(1, "can't specify both interface (-i) and capture file (-r)");
+
+   if ((hosts_max != 0) && (hosts_keep >= hosts_max))
+      errx(1, "must keep fewer hosts than --hosts-max (%u)", hosts_max);
+   verbosef("max %u hosts, cutting down to %u when exceeded",
+      hosts_max, hosts_keep);
+
+   if ((ports_max != 0) && (ports_keep >= ports_max))
+      errx(1, "must keep fewer ports than --ports-max (%u)", ports_max);
+   verbosef("max %u ports per host, cutting down to %u when exceeded",
+      ports_max, ports_keep);
+}
+
+static void
+run_from_capfile(void)
+{
+   graph_init();
+   hosts_db_init();
+   cap_from_file(capfile, filter);
+   cap_stop();
+   if (export_fn != NULL) db_export(export_fn);
+   hosts_db_free();
+   graph_free();
 }
 
 /* --- Program body --- */
@@ -260,11 +319,17 @@ int
 main(int argc, char **argv)
 {
    test_64order();
-   printf(PACKAGE_STRING " (built with libpcap %d.%d)\n",
-      PCAP_VERSION_MAJOR, PCAP_VERSION_MINOR);
-
    parse_cmdline(argc-1, argv+1);
 
+   if (capfile) {
+      /*
+       * This is very different from a regular run against a network
+       * interface.
+       */
+      run_from_capfile();
+      return 0;
+   }
+
    /* must verbosef() before first fork to init lock */
    verbosef("starting up");
    if (pid_fn) pidfile_create(chroot_dir, pid_fn, privdrop_user);
@@ -278,11 +343,9 @@ main(int argc, char **argv)
 
    /* do this first as it forks - minimize memory use */
    if (want_dns) dns_init(privdrop_user);
-
-   /* Need root privs for these: */
-   cap_init(interface, filter, want_promisc);
-   http_init(bindaddr, bindport, /*maxconn=*/ -1);
-   ncache_init(); /* don't need priv, but must do before chroot */
+   cap_init(interface, filter, want_promisc); /* needs root */
+   http_init(bindaddr, bindport, /*maxconn=*/ -1); /* low ports need root */
+   ncache_init(); /* must do before chroot() */
 
    privdrop(chroot_dir, privdrop_user);
 
@@ -336,7 +399,6 @@ main(int argc, char **argv)
             err(1, "select()");
       }
       else {
-         hosts_db_reduce();
          graph_rotate();
          cap_poll(&rs);
          dns_poll();
