@@ -42,6 +42,9 @@
 #  endif
 # endif
 #endif
+#ifndef ETHERTYPE_PPPOE
+#define ETHERTYPE_PPPOE 0x8864
+#endif
 
 #include <net/if.h> /* struct ifreq */
 #include <netinet/in_systm.h> /* n_long */
@@ -58,6 +61,8 @@ static void decode_ppp(u_char *, const struct pcap_pkthdr *,
    const u_char *);
 static void decode_pppoe(u_char *, const struct pcap_pkthdr *,
    const u_char *);
+static void decode_pppoe_real(const u_char *pdata, const uint32_t len,
+   pktsummary *sm);
 static void decode_linux_sll(u_char *, const struct pcap_pkthdr *,
    const u_char *);
 static void decode_ip(const u_char *pdata, const uint32_t len,
@@ -71,7 +76,7 @@ static const linkhdr_t linkhdrs[] = {
    { DLT_NULL,      NULL_HDR_LEN,  decode_loop  },
    { DLT_PPP,       PPP_HDR_LEN,   decode_ppp },
 #if defined(__NetBSD__)
-   { DLT_PPP_SERIAL, PPP_HDR_LEN, decode_ppp },
+   { DLT_PPP_SERIAL, PPP_HDR_LEN,  decode_ppp },
 #endif
    { DLT_FDDI,      FDDI_HDR_LEN,  NULL },
    { DLT_PPP_ETHER, PPPOE_HDR_LEN, decode_pppoe },
@@ -128,8 +133,10 @@ decode_ether(u_char *user _unused_,
 {
    u_short type;
    const struct ether_header *hdr = (const struct ether_header *)pdata;
+   extern int want_pppoe;
    pktsummary sm;
    memset(&sm, 0, sizeof(sm));
+   sm.time = pheader->ts.tv_sec;
 
    if (pheader->caplen < ETHER_HDR_LEN) {
       verbosef("ether: packet too short (%u bytes)", pheader->caplen);
@@ -147,15 +154,25 @@ decode_ether(u_char *user _unused_,
    type = ntohs( hdr->ether_type );
    switch (type) {
    case ETHERTYPE_IP:
-      decode_ip(pdata + ETHER_HDR_LEN, pheader->caplen - ETHER_HDR_LEN, &sm);
-      sm.time = pheader->ts.tv_sec;
-      acct_for(&sm);
+      if (!want_pppoe) {
+         decode_ip(pdata + ETHER_HDR_LEN,
+                   pheader->caplen - ETHER_HDR_LEN, &sm);
+         acct_for(&sm);
+      } else
+         verbosef("ether: discarded IP packet, expecting PPPoE instead");
       break;
    case ETHERTYPE_ARP:
       /* known protocol, don't complain about it. */
       break;
+   case ETHERTYPE_PPPOE:
+      if (want_pppoe)
+         decode_pppoe_real(pdata + ETHER_HDR_LEN,
+                           pheader->caplen - ETHER_HDR_LEN, &sm);
+      else
+         verbosef("ether: got PPPoE frame: maybe you want --pppoe");
+      break;
    default:
-      verbosef("ether: unknown protocol (%04x)", type);
+      verbosef("ether: unknown protocol (0x%04x)", type);
    }
 }
 
@@ -194,6 +211,11 @@ decode_ppp(u_char *user _unused_,
    pktsummary sm;
    memset(&sm, 0, sizeof(sm));
 
+   if (pheader->caplen < PPPOE_HDR_LEN) {
+      verbosef("ppp: packet too short (%u bytes)", pheader->caplen);
+      return;
+   }
+
    if (pdata[2] == 0x00 && pdata[3] == 0x21) {
          decode_ip(pdata + PPP_HDR_LEN, pheader->caplen - PPP_HDR_LEN, &sm);
          sm.time = pheader->ts.tv_sec;
@@ -209,32 +231,33 @@ decode_pppoe(u_char *user _unused_,
 {
    pktsummary sm;
    memset(&sm, 0, sizeof(sm));
+   sm.time = pheader->ts.tv_sec;
+   decode_pppoe_real(pdata, pheader->caplen, &sm);
+}
 
-   if (pheader->caplen < PPPOE_HDR_LEN) {
-      verbosef("loop: packet too short (%u bytes)", pheader->caplen);
+static void
+decode_pppoe_real(const u_char *pdata, const uint32_t len,
+   pktsummary *sm)
+{
+   if (len < PPPOE_HDR_LEN) {
+      verbosef("pppoe: packet too short (%u bytes)", len);
       return;
    }
 
-   /*
-    * First, check if this is a session PPPoE packet
-    * by checking the CODE part of the header
-    * (2nd byte in header)
-    */
-   if (pdata[1] == 0x00) {
-      /*
-       * Next, check if PPP packet type is 0x0021,
-       * which is PPP_IP
-       */
-      if (pdata[6] == 0x00 && pdata[7] == 0x21) {
-         decode_ip(pdata + PPPOE_HDR_LEN, pheader->caplen - PPPOE_HDR_LEN, &sm);
-         sm.time = pheader->ts.tv_sec;
-         acct_for(&sm);
-      }
-      else
-         verbosef("non-IP PPPoE packet; ignoring.");
+   if (pdata[1] != 0x00) {
+      verbosef("pppoe: code = 0x%02x, expecting 0; ignoring.", pdata[1]);
+      return;
    }
-   else
-      verbosef("non-session PPPoE packet; ignoring.");
+
+   if ((pdata[6] == 0xc0) && (pdata[7] == 0x21)) return; /* LCP */
+   if ((pdata[6] == 0xc0) && (pdata[7] == 0x25)) return; /* LQR */
+
+   if ((pdata[6] == 0x00) && (pdata[7] == 0x21)) {
+      decode_ip(pdata + PPPOE_HDR_LEN, len - PPPOE_HDR_LEN, sm);
+      acct_for(sm);
+   } else
+      verbosef("pppoe: non-IP PPPoE packet (0x%02x%02x); ignoring.",
+         pdata[6], pdata[7]);
 }
 
 /* very similar to decode_ether ... */
