@@ -4,6 +4,8 @@
  * graph_db.h: round robin database for graph data
  */
 
+#include <stdint.h> /* for uint64_t on Linux and OS X */
+
 enum graph_dir {
    MIN_GRAPH_DIR = 1,
    GRAPH_IN = 1,
@@ -12,9 +14,12 @@ enum graph_dir {
 };
 
 void graph_init(void);
+void graph_reset(void);
 void graph_free(void);
 void graph_acct(uint64_t amount, enum graph_dir dir);
 void graph_rotate(void);
+int graph_import(const int fd);
+int graph_export(const int fd);
 
 struct str *html_front_page(void);
 struct str *xml_graphs(void);
