@@ -1,5 +1,5 @@
 /* darkstat 3
- * copyright (c) 2001-2007 Emil Mikulic.
+ * copyright (c) 2001-2008 Emil Mikulic.
  *
  * hosts_db.c: database of hosts, ports, protocols.
  *
@@ -26,18 +26,13 @@
 #include <unistd.h>
 
 int show_mac_addrs = 0;
+extern const char *interface;
 
 /* FIXME: specify somewhere more sane/tunable */
 #define MAX_ENTRIES 30 /* in an HTML table rendered from a hashtable */
 
-/*
- * Hosts table reduction - when the number of hosts hits NUM_HOSTS_HIGH, we
- * reduce the table to NUM_HOSTS_LOW hosts.
- */
-#define NUM_HOSTS_HIGH 1000
-#define NUM_HOSTS_LOW  500
-
 typedef uint32_t (hash_func_t)(const struct hashtable *, const void *);
+typedef void (free_func_t)(struct bucket *);
 typedef const void * (key_func_t)(const struct bucket *);
 typedef int (find_func_t)(const struct bucket *, const void *);
 typedef struct bucket * (make_func_t)(const void *);
@@ -47,7 +42,7 @@ typedef void (format_row_func_t)(struct str *, const struct bucket *);
 struct hashtable {
    uint8_t bits;     /* size of hashtable in bits */
    uint32_t size, mask;
-   uint32_t count;   /* items in table */
+   uint32_t count, count_max, count_keep;   /* items in table */
    uint32_t coeff;   /* coefficient for Fibonacci hashing */
    struct bucket **table;
 
@@ -58,6 +53,9 @@ struct hashtable {
    hash_func_t *hash_func;
    /* returns hash value of given key (passed as void*) */
 
+   free_func_t *free_func;
+   /* free of bucket payload */
+
    key_func_t *key_func;
    /* returns pointer to key of bucket (to pass to hash_func) */
 
@@ -74,6 +72,9 @@ struct hashtable {
    /* format record and append to str */
 };
 
+static void hashtable_reduce(struct hashtable *ht);
+static void hashtable_free(struct hashtable *h);
+
 #define HOST_BITS 1  /* initial size of hosts table */
 #define PORT_BITS 1  /* initial size of ports tables */
 #define PROTO_BITS 1 /* initial size of proto table */
@@ -207,6 +208,16 @@ make_func_host(const void *key)
    return (b);
 }
 
+static void
+free_func_host(struct bucket *b)
+{
+   struct host *h = &(b->u.host);
+   if (h->dns != NULL) free(h->dns);
+   hashtable_free(h->ports_tcp);
+   hashtable_free(h->ports_udp);
+   hashtable_free(h->ip_protos);
+}
+
 static struct bucket *
 make_func_port_tcp(const void *key)
 {
@@ -232,6 +243,12 @@ make_func_ip_proto(const void *key)
    return (b);
 }
 
+static void
+free_func_simple(struct bucket *b _unused_)
+{
+   /* nop */
+}
+
 /* ---------------------------------------------------------------------------
  * format_func collection (ordered by struct)
  */
@@ -404,7 +421,10 @@ format_row_ip_proto(struct str *buf, const struct bucket *b)
  */
 static struct hashtable *
 hashtable_make(const uint8_t bits,
+   const unsigned int count_max,
+   const unsigned int count_keep,
    hash_func_t *hash_func,
+   free_func_t *free_func,
    key_func_t *key_func,
    find_func_t *find_func,
    make_func_t *make_func,
@@ -416,10 +436,13 @@ hashtable_make(const uint8_t bits,
 
    hash = xmalloc(sizeof(*hash));
    hash->bits = bits;
+   hash->count_max = count_max;
+   hash->count_keep = count_keep;
    hash->size = 1U << bits;
    hash->mask = hash->size - 1;
    hash->coeff = coprime(hash->size);
    hash->hash_func = hash_func;
+   hash->free_func = free_func;
    hash->key_func = key_func;
    hash->find_func = find_func;
    hash->make_func = make_func;
@@ -438,9 +461,9 @@ void
 hosts_db_init(void)
 {
    assert(hosts_db == NULL);
-   hosts_db = hashtable_make(HOST_BITS,
-      hash_func_host, key_func_host, find_func_host, make_func_host,
-      format_cols_host, format_row_host);
+   hosts_db = hashtable_make(HOST_BITS, hosts_max, hosts_keep,
+      hash_func_host, free_func_host, key_func_host, find_func_host,
+      make_func_host, format_cols_host, format_row_host);
 }
 
 static void
@@ -474,8 +497,6 @@ hashtable_rehash(struct hashtable *h, const uint8_t bits)
 
    free(h->table);
    h->table = new_table;
-   verbosef("hashtable rehashed to %d slots (was using %d of %d)",
-      h->size, h->count, old_size);
 }
 
 static void
@@ -529,7 +550,9 @@ hashtable_find_or_insert(struct hashtable *h, const void *key)
    struct bucket *b = hashtable_search(h, key);
 
    if (b == NULL) {
-      /* Not found so insert. */
+      /* Not found so insert.  But first check occupancy. */
+      assert(h->count <= h->count_max);
+      if (h->count == h->count_max) hashtable_reduce(h);
       b = h->make_func(key);
       hashtable_insert(h, b);
    }
@@ -542,7 +565,7 @@ hashtable_find_or_insert(struct hashtable *h, const void *key)
  * the bucket.
  */
 static void
-hashtable_free_simple(struct hashtable *h)
+hashtable_free(struct hashtable *h)
 {
    uint32_t i;
 
@@ -553,6 +576,7 @@ hashtable_free_simple(struct hashtable *h)
       while (b != NULL) {
          tmp = b;
          b = b->next;
+         h->free_func(tmp);
          free(tmp);
       }
    }
@@ -560,16 +584,6 @@ hashtable_free_simple(struct hashtable *h)
    free(h);
 }
 
-static void
-host_free(struct host *h)
-{
-   if (h->dns != NULL)
-      free(h->dns);
-   hashtable_free_simple(h->ports_tcp);
-   hashtable_free_simple(h->ports_udp);
-   hashtable_free_simple(h->ip_protos);
-}
-
 /* ---------------------------------------------------------------------------
  * Return existing host or insert a new one.
  */
@@ -602,57 +616,56 @@ host_search(const char *ipstr)
 }
 
 /* ---------------------------------------------------------------------------
- * Reduce hosts_db if necessary.
+ * Reduce a hashtable to the top <keep> entries.
  */
-void
-hosts_db_reduce(void)
+static void
+hashtable_reduce(struct hashtable *ht)
 {
    uint32_t i, pos, rmd;
    const struct bucket **table;
    uint64_t cutoff;
 
-   if (hosts_db->count < NUM_HOSTS_HIGH)
-      return;
+   assert(ht->count_keep < ht->count);
 
    /* Fill table with pointers to buckets in hashtable. */
-   table = xmalloc(sizeof(*table) * hosts_db->count);
-   for (pos=0, i=0; i<hosts_db->size; i++) {
-      struct bucket *b = hosts_db->table[i];
+   table = xmalloc(sizeof(*table) * ht->count);
+   for (pos=0, i=0; i<ht->size; i++) {
+      struct bucket *b = ht->table[i];
       while (b != NULL) {
          table[pos++] = b;
          b = b->next;
       }
    }
-   assert(pos == hosts_db->count);
-   qsort_buckets(table, hosts_db->count, 0, NUM_HOSTS_LOW, TOTAL);
-   cutoff = table[NUM_HOSTS_LOW - 1]->total;
+   assert(pos == ht->count);
+   qsort_buckets(table, ht->count, 0, ht->count_keep, TOTAL);
+   cutoff = table[ht->count_keep]->total;
    free(table);
 
-   /* Remove all hosts with total < cutoff. */
+   /* Remove all elements with total <= cutoff. */
    rmd = 0;
-   for (i=0; i<hosts_db->size; i++) {
-      struct bucket *last = NULL, *next, *b = hosts_db->table[i];
+   for (i=0; i<ht->size; i++) {
+      struct bucket *last = NULL, *next, *b = ht->table[i];
       while (b != NULL) {
          next = b->next;
-         if (b->total < cutoff) {
+         if (b->total <= cutoff) {
             /* Remove this one. */
-            host_free(&(b->u.host));
+            ht->free_func(b);
             free(b);
             if (last == NULL)
-               hosts_db->table[i] = next;
+               ht->table[i] = next;
             else
                last->next = next;
             rmd++;
-            hosts_db->count--;
+            ht->count--;
          } else {
             last = b;
          }
          b = next;
       }
    }
-   verbosef("hosts_db_reduce: from %u, removed %u hosts",
-      hosts_db->count+rmd, rmd);
-   hashtable_rehash(hosts_db, hosts_db->bits); /* needed? */
+   verbosef("hashtable_reduce: removed %u buckets, left %u",
+      rmd, ht->count);
+   hashtable_rehash(ht, ht->bits); /* is this needed? */
 }
 
 /* ---------------------------------------------------------------------------
@@ -667,7 +680,7 @@ hosts_db_reset(void)
       struct bucket *next, *b = hosts_db->table[i];
       while (b != NULL) {
          next = b->next;
-         host_free(&(b->u.host));
+         hosts_db->free_func(b);
          free(b);
          b = next;
       }
@@ -688,9 +701,9 @@ void hosts_db_free(void)
    for (i=0; i<hosts_db->size; i++) {
       struct bucket *tmp, *b = hosts_db->table[i];
       while (b != NULL) {
-         host_free(&(b->u.host));
          tmp = b;
          b = b->next;
+         hosts_db->free_func(tmp);
          free(tmp);
       }
    }
@@ -708,8 +721,8 @@ host_get_port_tcp(struct bucket *host, const uint16_t port)
    struct host *h = &host->u.host;
    assert(h != NULL);
    if (h->ports_tcp == NULL)
-      h->ports_tcp = hashtable_make(PORT_BITS,
-         hash_func_short, key_func_port_tcp,
+      h->ports_tcp = hashtable_make(PORT_BITS, ports_max, ports_keep,
+         hash_func_short, free_func_simple, key_func_port_tcp,
          find_func_port_tcp, make_func_port_tcp,
          format_cols_port_tcp, format_row_port_tcp);
    return (hashtable_find_or_insert(h->ports_tcp, &port));
@@ -724,8 +737,8 @@ host_get_port_udp(struct bucket *host, const uint16_t port)
    struct host *h = &host->u.host;
    assert(h != NULL);
    if (h->ports_udp == NULL)
-      h->ports_udp = hashtable_make(PORT_BITS,
-         hash_func_short, key_func_port_udp,
+      h->ports_udp = hashtable_make(PORT_BITS, ports_max, ports_keep,
+         hash_func_short, free_func_simple, key_func_port_udp,
          find_func_port_udp, make_func_port_udp,
          format_cols_port_udp, format_row_port_udp);
    return (hashtable_find_or_insert(h->ports_udp, &port));
@@ -740,8 +753,8 @@ host_get_ip_proto(struct bucket *host, const uint8_t proto)
    struct host *h = &host->u.host;
    assert(h != NULL);
    if (h->ip_protos == NULL)
-      h->ip_protos = hashtable_make(PROTO_BITS,
-         hash_func_byte, key_func_ip_proto,
+      h->ip_protos = hashtable_make(PROTO_BITS, ports_max, ports_keep,
+         hash_func_byte, free_func_simple, key_func_ip_proto,
          find_func_ip_proto, make_func_ip_proto,
          format_cols_ip_proto, format_row_ip_proto);
    return (hashtable_find_or_insert(h->ip_protos, &proto));
@@ -770,21 +783,9 @@ html_hosts(const char *uri, const char *query)
    if (num_elems == 1)
       /* /hosts/ */
       buf = html_hosts_main(query);
-   else if ((num_elems == 2) && (strcmp(elem[1], "stats") == 0))
-      /* /hosts/stats/ */
-      buf = html_hashtable_stats(hosts_db);
    else if (num_elems == 2)
       /* /hosts/<IP of host>/ */
       buf = html_hosts_detail(elem[1]);
-   else if ((num_elems == 3) && (strcmp(elem[2], "tcpstats") == 0))
-      /* /hosts/<IP of host>/tcpstats/ */
-      buf = html_tcp_stats(elem[1]);
-   else if ((num_elems == 3) && (strcmp(elem[2], "udpstats") == 0))
-      /* /hosts/<IP of host>/udpstats/ */
-      buf = html_udp_stats(elem[1]);
-   else if ((num_elems == 3) && (strcmp(elem[2], "protostats") == 0))
-      /* /hosts/<IP of host>/protostats/ */
-      buf = html_proto_stats(elem[1]);
 
    for (i=0; i<num_elems; i++)
       free(elem[i]);
@@ -887,9 +888,9 @@ html_hosts_main(const char *qs)
 #define FULL "full table"
 
    str_append(buf, html_header_1);
-   str_append(buf, "<title>darkstat3: Hosts</title>\n");
+   str_appendf(buf, "<title>darkstat3: Hosts (%s)</title>\n", interface);
    str_append(buf, html_header_2);
-   str_append(buf, "<h2 class=\"pageheader\">Hosts</h2>\n");
+   str_appendf(buf, "<h2 class=\"pageheader\">Hosts (%s)</h2>\n", interface);
    format_table(buf, hosts_db, start, sort, full);
 
    /* <prev | full | stats | next> */
@@ -908,13 +909,11 @@ html_hosts_main(const char *qs)
       str_appendf(buf, " | <a href=\"?full=yes&sort=%s\">" FULL "</a>",
          sortstr);
 
-   str_append(buf, " | <a href=\"stats/\">statistics</a> | ");
-
    if (start+MAX_ENTRIES < (int)hosts_db->count)
-      str_appendf(buf, "<a href=\"?start=%d&sort=%s\">" NEXT "</a>",
+      str_appendf(buf, "| <a href=\"?start=%d&sort=%s\">" NEXT "</a>",
          start+MAX_ENTRIES, sortstr);
    else
-      str_append(buf, NEXT);
+      str_append(buf, " | " NEXT);
 
    str_append(buf, "<br/>\n");
    str_append(buf, html_footer);
@@ -974,124 +973,20 @@ html_hosts_detail(const char *ip)
 
    str_append(buf, "<h3>TCP ports</h3>\n");
    format_table(buf, h->u.host.ports_tcp, 0,TOTAL,0);
-   str_append(buf, "<a href=\"tcpstats/\">statistics</a><br/>\n");
 
    str_append(buf, "<h3>UDP ports</h3>\n");
    format_table(buf, h->u.host.ports_udp, 0,TOTAL,0);
-   str_append(buf, "<a href=\"udpstats/\">statistics</a><br/>\n");
 
    str_append(buf, "<h3>IP protocols</h3>\n");
    format_table(buf, h->u.host.ip_protos, 0,TOTAL,0);
-   str_append(buf, "<a href=\"protostats/\">statistics</a><br/>\n");
 
    str_append(buf, html_footer);
    return (buf);
 }
 
-/* ---------------------------------------------------------------------------
- * Web interface: statistics of a given hashtable.
- */
-static struct str *
-html_hashtable_stats(struct hashtable *h)
-{
-   struct str *buf;
-   uint32_t i, max_len = 0, num_empty = 0, ovr_buckets = 0, ovr_slots = 0;
-
-   buf = str_make();
-   str_appendf(buf,
-      "%s"
-      " <title>Hashtable histogram</title>\n"
-      " <style type=\"text/css\">\n"
-      "  thead td { font-weight:bold; background:#E0E0E0; }\n"
-      "  tbody td { text-align:right; background:#F0F0F0; }\n"
-      " </style>\n"
-      "%s"
-      "<h2>Hashtable statistics</h2>\n",
-      html_header_1, html_header_2
-   );
-
-   if (h == NULL)
-      str_append(buf, "<p>Table is empty.</p>\n");
-   else {
-      /* Traverse table. */
-      for (i=0; i<h->size; i++)
-      {
-         struct bucket *b = h->table[i];
-         uint32_t curr_len = 0;
-
-         if (b == NULL)
-            num_empty++;
-         else
-            for (b = h->table[i]; b != NULL; b = b->next)
-               curr_len++;
-
-         if (curr_len > 1)
-         {
-            ovr_slots++;
-            ovr_buckets += curr_len - 1;
-            max_len = max(max_len, curr_len);
-         }
-      }
-
-      str_appendf(buf,
-         "<p>\n"
-         "<b>Slots:</b> %'u<br/>\n"
-         "<b>Records:</b> %'u<br/>\n"
-         "<b>Slots empty:</b> %'u<br/>\n"
-         "<b>Slots overflowed:</b> %'u<br/>\n"
-         "<b>Overflow buckets:</b> %'u<br/>\n"
-         "<b>Longest chain:</b> %'u<br/>\n"
-         "<b>Usage:</b> %u%%<br/>\n"
-         "<br/>\n"
-         "<b>Inserts:</b> %'qu<br/>\n"
-         "<b>Searches:</b> %'qu<br/>\n"
-         "<b>Deletions:</b> %'qu<br/>\n"
-         "<b>Rehashes:</b> %'qu<br/>\n"
-         "</p>\n"
-         ,
-         h->size, h->count, num_empty, ovr_slots, ovr_buckets, max_len,
-         100 * h->count / h->size,
-         h->stats.inserts, h->stats.searches,
-         h->stats.deletions, h->stats.rehashes);
-   }
-
-   str_append(buf, html_footer);
-   return (buf);
-}
-
-static struct str *
-html_tcp_stats(const char *ip)
-{
-   struct bucket *h = host_search(ip);
-
-   if (h == NULL)
-      return (NULL); /* no such host */
-   return html_hashtable_stats(h->u.host.ports_tcp);
-}
-
-static struct str *
-html_udp_stats(const char *ip)
-{
-   struct bucket *h = host_search(ip);
-
-   if (h == NULL)
-      return (NULL); /* no such host */
-   return html_hashtable_stats(h->u.host.ports_udp);
-}
-
-static struct str *
-html_proto_stats(const char *ip)
-{
-   struct bucket *h = host_search(ip);
-
-   if (h == NULL)
-      return (NULL); /* no such host */
-   return html_hashtable_stats(h->u.host.ip_protos);
-}
-
 /* ---------------------------------------------------------------------------
  * Database import and export code:
- * Initially written and contributed by Ben stewart.
+ * Initially written and contributed by Ben Stewart.
  * copyright (c) 2007 Ben Stewart, Emil Mikulic.
  */
 static int hosts_db_export_ip(const struct hashtable *h, const int fd);
