@@ -1,5 +1,5 @@
 /* darkstat 3
- * copyright (c) 2001-2006 Emil Mikulic.
+ * copyright (c) 2001-2008 Emil Mikulic.
  *
  * http.c: embedded webserver.
  * This borrows a lot of code from darkhttpd.
@@ -15,7 +15,9 @@
 #include "graph_db.h"
 #include "err.h"
 #include "queue.h"
+#include "now.h"
 
+#include <sys/uio.h>
 #include <sys/socket.h>
 #include <arpa/inet.h>
 #include <assert.h>
@@ -27,7 +29,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <time.h>
 #include <unistd.h>
 #include <zlib.h>
 
@@ -35,9 +36,9 @@ static const char mime_type_xml[] = "text/xml";
 static const char mime_type_html[] = "text/html; charset=us-ascii";
 static const char mime_type_css[] = "text/css";
 static const char mime_type_js[] = "text/javascript";
-static const char encoding_deflate[] =
+static const char encoding_gzip[] =
     "Vary: Accept-Encoding\r\n"
-    "Content-Encoding: deflate\r\n";
+    "Content-Encoding: gzip\r\n";
 
 static const char server[] = PACKAGE_NAME "/" PACKAGE_VERSION;
 static int idletime = 60;
@@ -55,19 +56,20 @@ struct connection {
     in_addr_t client;
     time_t last_active;
     enum {
-        RECV_REQUEST,   /* receiving request */
-        SEND_HEADER,    /* sending generated header */
-        SEND_REPLY,     /* sending reply */
-        DONE            /* connection closed, need to remove from queue */
+        RECV_REQUEST,          /* receiving request */
+        SEND_HEADER_AND_REPLY, /* try to send header+reply together */
+        SEND_HEADER,           /* sending generated header */
+        SEND_REPLY,            /* sending reply */
+        DONE                   /* conn closed, need to remove from queue */
         } state;
 
     /* char request[request_length+1] is null-terminated */
     char *request;
     size_t request_length;
-    int accept_deflate;
+    int accept_gzip;
 
     /* request fields */
-    char *method, *uri;
+    char *method, *uri, *query; /* query can be NULL */
 
     char *header;
     const char *mime_type, *encoding, *header_extra;
@@ -256,12 +258,13 @@ static struct connection *new_connection(void)
 
     conn->socket = -1;
     conn->client = INADDR_ANY;
-    conn->last_active = time(NULL);
+    conn->last_active = now;
     conn->request = NULL;
     conn->request_length = 0;
-    conn->accept_deflate = 0;
+    conn->accept_gzip = 0;
     conn->method = NULL;
     conn->uri = NULL;
+    conn->query = NULL;
     conn->header = NULL;
     conn->mime_type = NULL;
     conn->encoding = "";
@@ -335,6 +338,7 @@ static void free_connection(struct connection *conn)
     if (conn->request != NULL) free(conn->request);
     if (conn->method != NULL) free(conn->method);
     if (conn->uri != NULL) free(conn->uri);
+    if (conn->query != NULL) free(conn->query);
     if (conn->header != NULL && !conn->header_dont_free) free(conn->header);
     if (conn->reply != NULL && !conn->reply_dont_free) free(conn->reply);
 }
@@ -348,9 +352,9 @@ static void free_connection(struct connection *conn)
 #define DATE_LEN 30 /* strlen("Fri, 28 Feb 2003 00:02:08 GMT")+1 */
 static char *rfc1123_date(char *dest, const time_t when)
 {
-    time_t now = when;
+    time_t tmp = when;
     if (strftime(dest, DATE_LEN,
-        "%a, %d %b %Y %H:%M:%S %Z", gmtime(&now) ) == 0)
+        "%a, %d %b %Y %H:%M:%S %Z", gmtime(&tmp) ) == 0)
             errx(1, "strftime() failed [%s]", dest);
     return (dest);
 }
@@ -371,7 +375,7 @@ static void default_reply(struct connection *conn,
     va_end(va);
 
     /* Only really need to calculate the date once. */
-    (void)rfc1123_date(date, time(NULL));
+    (void)rfc1123_date(date, now);
 
     conn->reply_length = xasprintf(&(conn->reply),
      "<html><head><title>%d %s</title></head><body>\n"
@@ -429,13 +433,14 @@ static char *parse_field(const struct connection *conn, const char *field)
 
 
 /* ---------------------------------------------------------------------------
- * Parse an HTTP request like "GET / HTTP/1.1" to get the method (GET), the
- * url (/), and if the UA will accept deflate encoding.  Remember to
- * deallocate all these buffers.  The method will be returned in uppercase.
+ * Parse an HTTP request like "GET /hosts/?sort=in HTTP/1.1" to get the method
+ * (GET), the uri (/hosts/), the query (sort=in) and whether the UA will
+ * accept gzip encoding.  Remember to deallocate all these buffers.  Query
+ * can be NULL.  The method will be returned in uppercase.
  */
 static int parse_request(struct connection *conn)
 {
-    size_t bound1, bound2;
+    size_t bound1, bound2, mid;
     char *accept_enc;
 
     /* parse method */
@@ -459,13 +464,22 @@ static int parse_request(struct connection *conn)
         conn->request[bound2] != '\r'; bound2++)
             ;
 
+    /* find query string */
+    for (mid=bound1; mid<bound2 && conn->request[mid] != '?'; mid++)
+        ;
+
+    if (conn->request[mid] == '?') {
+        conn->query = split_string(conn->request, mid+1, bound2);
+        bound2 = mid;
+    }
+
     conn->uri = split_string(conn->request, bound1, bound2);
 
     /* parse important fields */
     accept_enc = parse_field(conn, "Accept-Encoding: ");
     if (accept_enc != NULL) {
-        if (strstr(accept_enc, "deflate") != NULL)
-            conn->accept_deflate = 1;
+        if (strstr(accept_enc, "gzip") != NULL)
+            conn->accept_gzip = 1;
         free(accept_enc);
     }
     return (1);
@@ -502,17 +516,17 @@ static_graph_js(struct connection *conn)
 }
 
 /* ---------------------------------------------------------------------------
- * Deflate a reply, if requested and possible.  Don't bother with a minimum
- * length requirement since I've never seen a page fail to compress.
+ * gzip a reply, if requested and possible.  Don't bother with a minimum
+ * length requirement, I've never seen a page fail to compress.
  */
 static void
-process_deflate(struct connection *conn)
+process_gzip(struct connection *conn)
 {
     char *buf;
     size_t len;
     z_stream zs;
 
-    if (!conn->accept_deflate)
+    if (!conn->accept_gzip)
         return;
 
     buf = xmalloc(conn->reply_length);
@@ -522,7 +536,10 @@ process_deflate(struct connection *conn)
     zs.zfree = Z_NULL;
     zs.opaque = Z_NULL;
 
-    if (deflateInit(&zs, Z_BEST_COMPRESSION) != Z_OK)
+    if (deflateInit2(&zs, Z_BEST_COMPRESSION, Z_DEFLATED,
+        15+16, /* 15 = biggest window, 16 = add gzip header+trailer */
+        8 /* default */,
+        Z_DEFAULT_STRATEGY) != Z_OK)
        return;
 
     zs.avail_in = conn->reply_length;
@@ -545,17 +562,10 @@ process_deflate(struct connection *conn)
 
     conn->reply = buf;
     conn->reply_length -= zs.avail_out;
-    conn->encoding = encoding_deflate;
+    conn->encoding = encoding_gzip;
     deflateEnd(&zs);
     verbosef("compressed %u bytes to %u bytes",
        (unsigned int)len, (unsigned int)conn->reply_length);
-
-    /* To unconfuse IE (and apparently Safari), we must trim 2 bytes from
-     * the header.  It's also been suggested to trim 5 from the footer
-     * but I've found that this breaks things.
-     */
-    conn->reply_sent = 2;
-    conn->reply_length -= 2;
 }
 
 /* ---------------------------------------------------------------------------
@@ -566,7 +576,8 @@ static void process_get(struct connection *conn)
     char *decoded_url, *safe_url;
     char date[DATE_LEN];
 
-    verbosef("http: [%s %s]", conn->method, conn->uri);
+    verbosef("http: [%s %s [%s]]", conn->method, conn->uri,
+        (conn->query == NULL)?"<null>":conn->query);
 
     /* work out path of file being requested */
     decoded_url = urldecode(conn->uri);
@@ -588,7 +599,7 @@ static void process_get(struct connection *conn)
     }
     else if (str_starts_with(safe_url, "/hosts/")) {
         /* FIXME here - make this saner */
-        struct str *buf = html_hosts(safe_url);
+        struct str *buf = html_hosts(safe_url, conn->query);
         if (buf == NULL) {
             default_reply(conn, 404, "Not Found",
                 "The page you requested could not be found.");
@@ -615,7 +626,7 @@ static void process_get(struct connection *conn)
     }
     free(safe_url);
 
-    process_deflate(conn);
+    process_gzip(conn);
     assert(conn->mime_type != NULL);
     conn->header_length = xasprintf(&(conn->header),
         "HTTP/1.1 200 OK\r\n"
@@ -627,7 +638,7 @@ static void process_get(struct connection *conn)
         "%s"
         "\r\n"
         ,
-        rfc1123_date(date, time(NULL)), server,
+        rfc1123_date(date, now), server,
         conn->reply_length, conn->mime_type, conn->encoding,
         conn->header_extra);
     conn->http_code = 200;
@@ -662,7 +673,10 @@ static void process_request(struct connection *conn)
     }
 
     /* advance state */
-    conn->state = SEND_HEADER;
+    if (conn->header_only)
+        conn->state = SEND_HEADER;
+    else
+        conn->state = SEND_HEADER_AND_REPLY;
 
     /* request not needed anymore */
     free(conn->request);
@@ -689,7 +703,7 @@ static void poll_recv_request(struct connection *conn)
         conn->state = DONE;
         return;
     }
-    conn->last_active = time(NULL);
+    conn->last_active = now;
     #undef BUFSIZE
 
     /* append to conn->request */
@@ -714,6 +728,65 @@ static void poll_recv_request(struct connection *conn)
 
 
 
+/* ---------------------------------------------------------------------------
+ * Try to send header and [a part of the] reply in one packet.
+ */
+static void poll_send_header_and_reply(struct connection *conn)
+{
+    ssize_t sent;
+    struct iovec iov[2];
+
+    assert(!conn->header_only);
+    assert(conn->reply_length > 0);
+    assert(conn->header_sent == 0);
+
+    assert(conn->reply_sent == 0);
+
+    /* Fill out iovec */
+    iov[0].iov_base = conn->header;
+    iov[0].iov_len = conn->header_length;
+
+    iov[1].iov_base = conn->reply + conn->reply_sent;
+    iov[1].iov_len = conn->reply_length - conn->reply_sent;
+
+    sent = writev(conn->socket, iov, 2);
+    conn->last_active = now;
+    verbosef("poll_send_header_and_reply(%d) sent %d bytes",
+        conn->socket, (int)sent);
+
+    /* handle any errors (-1) or closure (0) in send() */
+    if (sent < 1) {
+        if (sent == -1)
+            verbosef("writev(%d) error: %s", conn->socket, strerror(errno));
+        conn->state = DONE;
+        return;
+    }
+
+    /* Figure out what we've sent. */
+    conn->total_sent += (unsigned int)sent;
+    if (sent < (ssize_t)conn->header_length) {
+        verbosef("partially sent header");
+        conn->header_sent = sent;
+        conn->state = SEND_HEADER;
+        return;
+    }
+    /* else */
+    verbosef("completely sent header");
+    conn->header_sent = conn->header_length;
+    sent -= conn->header_length;
+
+    if (conn->reply_sent + sent < conn->reply_length) {
+        verbosef("partially sent reply");
+        conn->reply_sent += sent;
+        conn->state = SEND_REPLY;
+        return;
+    }
+    /* else */
+    verbosef("completely sent reply");
+    conn->reply_sent = conn->reply_length;
+    conn->state = DONE;
+}
+
 /* ---------------------------------------------------------------------------
  * Sending header.  Assumes conn->header is not NULL.
  */
@@ -721,11 +794,9 @@ static void poll_send_header(struct connection *conn)
 {
     ssize_t sent;
 
-    assert(conn->header_length == strlen(conn->header));
-
     sent = send(conn->socket, conn->header + conn->header_sent,
         conn->header_length - conn->header_sent, 0);
-    conn->last_active = time(NULL);
+    conn->last_active = now;
     dverbosef("poll_send_header(%d) sent %d bytes", conn->socket, (int)sent);
 
     /* handle any errors (-1) or closure (0) in send() */
@@ -761,7 +832,7 @@ static void poll_send_reply(struct connection *conn)
     sent = send(conn->socket,
         conn->reply + conn->reply_sent,
         conn->reply_length - conn->reply_sent, 0);
-    conn->last_active = time(NULL);
+    conn->last_active = now;
     dverbosef("poll_send_reply(%d) sent %d: [%d-%d] of %d",
         conn->socket, (int)sent,
         (int)conn->reply_sent,
@@ -837,7 +908,6 @@ http_fd_set(fd_set *recv_set, fd_set *send_set, int *max_fd,
     struct timeval *timeout, int *need_timeout)
 {
     struct connection *conn, *next;
-    time_t now = time(NULL);
     int minidle = idletime + 1;
 
     #define MAX_FD_SET(sock, fdset) do { \
@@ -873,6 +943,7 @@ http_fd_set(fd_set *recv_set, fd_set *send_set, int *max_fd,
             MAX_FD_SET(conn->socket, recv_set);
             break;
 
+        case SEND_HEADER_AND_REPLY:
         case SEND_HEADER:
         case SEND_REPLY:
             MAX_FD_SET(conn->socket, send_set);
@@ -909,6 +980,10 @@ void http_poll(fd_set *recv_set, fd_set *send_set)
         if (FD_ISSET(conn->socket, recv_set)) poll_recv_request(conn);
         break;
 
+    case SEND_HEADER_AND_REPLY:
+        if (FD_ISSET(conn->socket, send_set)) poll_send_header_and_reply(conn);
+        break;
+
     case SEND_HEADER:
         if (FD_ISSET(conn->socket, send_set)) poll_send_header(conn);
         break;
@@ -921,6 +996,4 @@ void http_poll(fd_set *recv_set, fd_set *send_set)
     }
 }
 
-
-
-/* vim:set tabstop=4 shiftwidth=4 expandtab tw=78: */
+/* vim:set ts=4 sw=4 et tw=78: */
